/*
 * (c) Copyright 2001-2003
 * Peter Blum, Peter.Blum@readme1st.de
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 *
 * Version: 2.1.0.0 - Aug 25 2003 11:11:11
 */

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <sys/types.h>

#include "s68.h"
#include "device.h"
#include "conf.h"



/*
 * d r a m R e s e t
 *
 * Reset the device
 */
int dramReset(Device *dev)
{
    if (!quiet)  uiPrintf(" DRAM");
    return 0;
}



/*
 * d r a m R e a d
 *
 * Read one byte
 */
static u_long dramRead(Device *dev, u_long offset)
{
    u_char *ram = (u_char *) dev->priv;

    return (u_long) ram[offset];
}



/*
 * d r a m W r i t e
 *
 * Write one byte
 */
static void dramWrite(Device *dev, u_long offset, u_char data)
{
    u_char *ram = (u_char *) dev->priv;

    ram[offset] = data;
}



/*
 * i n i t D R A M
 *
 * Initialize the device
 */
int initDRAM(Device *dev, u_long base, u_long size, void *arg)
{
    u_char  *ram;

    if ((ram = malloc(size)) == NULL) {
        uiErrPrintf("DRAM: Malloc of %ld bytes failed\n", size);
        return(errno);
    }

    dev->base = (u_long) base;
    dev->mask = ~(size-1);
    dev->priv = (void *) ram;
    dev->devReset = dramReset;
    dev->devRead  = dramRead;
    dev->devWrite = dramWrite;
    dev->devIAck  = NULL;
    dev->devCycle = NULL;

    return 0;
}
