/*
 * (c) Copyright 2001-2003
 * Peter Blum, Peter.Blum@readme1st.de
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 *
 * Version: 2.1.0.2 - May 02 2004 20:59:11
 *
 * 1  01.05.2004  Added macro for MSB calculation from size
 * 2  02.05.3004  Added macro for MASK calculation from size
 */

#ifndef _M68000_H
#define _M68000_H

/* Size of operand */
#define	 SZ_BYTE        0x000000ff      /*  8 bit */
#define	 SZ_WORD        0x0000ffff      /* 16 bit */
#define	 SZ_LONG        0xffffffff      /* 32 bit */

/* Get MSB for a given size */
#define  SZ_MSB(sz)    ((((u_long)sz) >> 1) + 1)
#define  SZ_MASK(sz)   ((u_long)sz)
/* CPU states */
#define  ST_RESET       0
#define  ST_RUNNING     1
#define  ST_STOP        2
#define  ST_HALT        3

/* Effective addressing */
#define  DATA_ADDR      0x0FFD
#define  MEMORY_ADDR    0x0FFC
#define  CONTROL_ADDR   0x07E4
#define  ALTERABLE_ADDR 0x01FF
#define  ALL_ADDR       0x0FFF
#define  DATA_ALT_ADDR  (DATA_ADDR & ALTERABLE_ADDR)
#define  MEM_ALT_ADDR   (MEMORY_ADDR & ALTERABLE_ADDR)
#define  CONT_ALT_ADDR  (CONTROL_ADDR & ALTERABLE_ADDR)

#define	 MODE_MASK      0x0038
#define	 REG_MASK       0x0007

#define	 READ           0xFFFF
#define	 WRITE          0x0000

/* Status register bitmasks */
#define  bit_1		0x0001
#define  bit_2		0x0002
#define  bit_3		0x0004
#define  bit_4		0x0008
#define  bit_5		0x0010
#define  bit_6		0x0020
#define  bit_7		0x0040
#define  bit_8		0x0080
#define  bit_9		0x0100
#define  bit_10		0x0200
#define  bit_11		0x0400
#define  bit_12		0x0800

#define	cbit		0x0001
#define	vbit		0x0002
#define	zbit		0x0004
#define	nbit		0x0008
#define	xbit		0x0010
#define	intmsk		0x0700         /* three bits */
#define	sbit		0x2000
#define	tbit		0x8000


/* Instruction return codes */
#define SUCCESS			0x0000
#define SUCCESS_PRIV		0x0001
#define BUS_ERROR		0x0002
#define ADDRESS_ERROR		0x0003
#define ILLEGAL_INST		0x0004
#define DIV_BY_ZERO		0x0005
#define CHK_EXCEPTION		0x0006
#define TRAPV_TRAP		0x0007
#define NO_PRIVILEGE		0x0008
#define TRAP_TRACE              0x0009
#define LINE_1010               0x000A
#define LINE_1111               0x000B
#define TRAP_TRAP		0x0010



/* Condition code setting */
#define	N_A		0
#define	GEN		1
#define	ZER		2
#define	UND		3
#define	CASE_1		4
#define	CASE_2		5
#define	CASE_3		6
#define	CASE_4		7
#define	CASE_5		8
#define	CASE_6		9
#define	CASE_7		10
#define	CASE_8		11
#define	CASE_9		12

/* Conditions for BCC, DBCC, and SCC */
#define T	0x00
#define F	0x01
#define HI	0x02
#define LS	0x03
#define CC	0x04
#define CS	0x05
#define NE	0x06
#define EQ	0x07
#define VC	0x08
#define VS	0x09
#define PL	0x0a
#define MI	0x0b
#define GE	0x0c
#define LT	0x0d
#define GT	0x0e
#define LE	0x0f


/* Processor register */
extern long  D[], A[], PC, SR;
#define  USP  A[7]
#define  SSP  A[8]

/* Return 8 for A7 in supervisor mode */
/* #define  a_reg(rn)   ((((rn) == 7) && (SR & sbit)) ? 8 : (rn)) */


extern int  trace;

int  cpuInit(void);
int  cpuReset(void);
void cpuExec(void);
void cpuSetIPL(int level);

extern int memPut(long data, long addr, long size);
extern int memGet(long addr, long size, long *result);
extern int memRequest(long *addrPtr, long size, long *result);

#endif

