/*
 *  Copyright (C) 1997-2001 B. Cameron Lesiuk
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Version: 2.1.0.0 - Aug 25 2003 11:11:11
 */

#include <windows.h>
#include "termbuf.h"



TBTERMBUF *tbAllocTermBuf(unsigned long p_ID, long bufX, long bufY, long scrX, long scrY)
{
	TBTERMBUF     *tBuf;
	unsigned long *lBuf;
	HGLOBAL       memory, bufmemory;
	long          x, y;
	unsigned long data;
	int           i;

	/* Allocate memory for TermBuf */
	memory = GlobalAlloc(GHND|GMEM_NOCOMPACT, sizeof(TBTERMBUF));
	if (!memory)
		return NULL;
	tBuf = (TBTERMBUF*)GlobalLock(memory);
	if (!tBuf) {
		GlobalFree(memory);
		return NULL;
	}

	bufmemory = GlobalAlloc(GHND|GMEM_NOCOMPACT, sizeof(long) * bufX * (bufY + 2) + 1024);
	if (!bufmemory) {
		GlobalUnlock(memory);
		GlobalFree(memory);
		return NULL;
	}
	lBuf=(unsigned long *)GlobalLock(bufmemory);
	if (!lBuf) {
		GlobalUnlock(memory);
		GlobalFree(memory);
		GlobalFree(bufmemory);
		return NULL;
	}
	tBuf->tMemory = memory;
	tBuf->tBufferMemory = bufmemory;
	tBuf->tBuffer = lBuf;
	tBuf->tTab  = &(lBuf[bufX * bufY]); /* tab stops are kept in an "extra" line at the end of the buffer */
	tBuf->tID   = p_ID;
	tBuf->tTopY = 0;
	tBuf->tBufX = bufX;
	tBuf->tBufY = bufY;
	tBuf->tCurX = 0;
	tBuf->tCurY = 0;
	tBuf->tScrX = scrX;
	tBuf->tScrY = scrY;
	tBuf->tScrollRegionActive = 0;
	tBuf->tSRX1 = 0;
	tBuf->tSRX2 = scrX;
	tBuf->tSRY1 = 0;
	tBuf->tSRY2 = scrY;
	tBuf->tCurrentAttribute = TB_MAKEFCOLOR(15L)|TB_MAKEBCOLOR(0L); /* white on black */
	tBuf->tHLStartX = tBuf->tHLStartY = 0L;
	tBuf->tHLEndX   = tBuf->tHLEndY   = 0L;
	tBuf->tHTab = 8;
	tBuf->tNotifyFn = NULL;
	tBuf->tAutoRightMargin = TRUE;

	/* Create default RGB color map */
	(tBuf->tColorBank)[ 0] = RGB(  0,  0,  0);
	(tBuf->tColorBank)[ 1] = RGB(128,  0,  0);
	(tBuf->tColorBank)[ 2] = RGB(  0,128,  0);
	(tBuf->tColorBank)[ 3] = RGB(128,128,  0);
	(tBuf->tColorBank)[ 4] = RGB(  0,  0,128);
	(tBuf->tColorBank)[ 5] = RGB(128,  0,128);
	(tBuf->tColorBank)[ 6] = RGB(  0,128,128);
	(tBuf->tColorBank)[ 7] = RGB(192,192,192);
	(tBuf->tColorBank)[ 8] = RGB(128,128,128);
	(tBuf->tColorBank)[ 9] = RGB(255,  0,  0);
	(tBuf->tColorBank)[10] = RGB(  0,255,  0);
	(tBuf->tColorBank)[11] = RGB(255,255,  0);
	(tBuf->tColorBank)[12] = RGB(  0,  0,255);
	(tBuf->tColorBank)[13] = RGB(255,  0,255);
	(tBuf->tColorBank)[14] = RGB(  0,255,255);
	(tBuf->tColorBank)[15] = RGB(255,255,255);
	for (i = 0; i < (1<<TB_COLORDEPTH); i++) {
		(tBuf->tColorPen)[i] = NULL;
	}

	/* Clear the buffer */
	data = TB_MAKEEFFECT(0)|TB_MAKEBCOLOR(0)|TB_MAKEFCOLOR(7)|TB_MAKECHAR(0x20);
	for (y = 0; y < bufY; y++)
		for (x = 0; x < bufX; x++)
			*(lBuf + x + (y * bufX)) = data;

	/* Reset tabs to the default - every 8 characters */
	for (i = 0; i < bufX; i++) {
		if (i%8)
			(tBuf->tTab)[i] = TBTAB_NONE;
		else
			(tBuf->tTab)[i] = TBTAB_LEFTALIGN;
	}

	return tBuf;
}


void tbFreeTermBuf(TBTERMBUF *termBuf)
{
	HGLOBAL  memory;
	int      i;

	if (!termBuf)  return;

	for (i = 0; i < (1<<TB_COLORDEPTH); i++) {
		if ((termBuf->tColorPen)[i]) {
			DeleteObject((termBuf->tColorPen)[i]);
		}
	}
	memory = termBuf->tBufferMemory;
	GlobalUnlock(memory);
	GlobalFree(memory);
	memory = termBuf->tMemory;
	GlobalUnlock(memory);
	GlobalFree(memory);
}


TBTERMBUF *tbReAllocTermBuf(TBTERMBUF *termBuf,long pBufX,long pBufY,long pScrX,long pScrY) {
	HGLOBAL oldmem, newmem;
	unsigned long *l_new, *t;
	long   x, y;
	long   spanX, spanY;
	long   xSrc, xDst, ySrc, yDst;
	unsigned long data;

	if (!termBuf)
		return NULL;

	if (pBufX < pScrX)  pBufX = pScrX;
	if (pBufY < pScrY)  pBufY = pScrY;

	if ((pBufX != termBuf->tBufX) || (pBufY != termBuf->tBufY)) {
		newmem = GlobalAlloc(GHND|GMEM_NOCOMPACT, sizeof(long) * pBufX * (pBufY+2)+1024);
		if (!newmem) {
			return termBuf;
		}
		l_new = (unsigned long *)GlobalLock(newmem);
		if (!l_new) {
			GlobalFree(newmem);
			return termBuf;
		}

		/* Clear the new buffer */
		data = TB_MAKEEFFECT(0)|TB_MAKEBCOLOR(0)|TB_MAKEFCOLOR(7)|TB_MAKECHAR(0x20);
		for (y = 0; y < (pBufY+2); y++)
			for (x = 0; x < pBufX; x++)
				*(l_new + x + (y*pBufX)) = data;
		/* Clear all tabs */
		for (x = 0; x < pBufX; x++)
			l_new[pBufX * pBufY+x] = TBTAB_NONE;

		/* Copy data over from old buffer to new one */
		spanX = TB_MIN(termBuf->tBufX, pBufX);
		spanY = TB_MIN(termBuf->tBufY, pBufY);
		ySrc = termBuf->tScrY-1;
		yDst = pScrY-1;
		for (y = 0; y < spanY; y++, ySrc--, yDst--) {
			if (yDst < 0) {
				t = l_new + ((pBufY + yDst) * pBufX);
			} else {
				t = l_new + (yDst * pBufX);
			}
			xSrc = xDst = 0;
			for (x = 0; x < spanX; x++, xSrc++, xDst++) {
				data = TB_GETCHAR(termBuf, xSrc, ySrc);
				*(t + xDst) = data;
			}
		}
		/* Reset tabs to the default - every 8 characters */
		for (x = 0; x < pBufX; x++) {
			if (x%8)
				l_new[pBufX * pBufY + x] = TBTAB_NONE;
			else
				l_new[pBufX * pBufY + x] = TBTAB_LEFTALIGN;
		}

		/* Copy over tabs from old buffer to new one */
		for (x = 0; x < spanX; x++) {
			l_new[pBufX * pBufY + x] = termBuf->tTab[x];
		}
		oldmem = termBuf->tBufferMemory;
		/* Swap in the new settings */
		termBuf->tBufferMemory = newmem;
		termBuf->tBuffer = l_new;
		termBuf->tTab = &(l_new[pBufX * pBufY]);
		termBuf->tTopY = 0;
		termBuf->tBufX = pBufX;
		termBuf->tBufY = pBufY;

		GlobalUnlock(oldmem);
		GlobalFree(oldmem);
	}
	/* Adjust the cursor too; it's top-relative */
	termBuf->tCurY -= termBuf->tScrY - pScrY;

	termBuf->tScrX = pScrX;
	termBuf->tScrY = pScrY;
	if (termBuf->tCurX >= pScrX)
		termBuf->tCurX = pScrX - 1;
	else if (termBuf->tCurX < 0)
		termBuf->tCurX = 0;
	if (termBuf->tCurY >= pScrY)
		termBuf->tCurY = pScrY-1;
	else if (termBuf->tCurY < 0)
		termBuf->tCurY = 0;
	termBuf->tHLStartX = termBuf->tHLStartY = 0;
	termBuf->tHLEndX   = termBuf->tHLEndY   = 0;

	return termBuf;
}


/*
 * tbPutChar writes the specified ASCII 8-bit character to the
 * terminal as would be expected, and does all attribute-assignment
 * and scrolling as required.
 */
void tbPutChar(TBTERMBUF *termBuf, unsigned char p_char)
{
	if (termBuf) {
		if (termBuf->tCurX > TB_GETXMAX(termBuf)) {
			/* we are off the end of our line. Throw in a wrap-around here first */
			termBuf->tCurX = 0;
			/* and do a newline */
			tbLineFeed(termBuf);
		}
		*TB_GETCHARADDR(termBuf, termBuf->tCurX, termBuf->tCurY) =
			termBuf->tCurrentAttribute|TB_MAKECHAR(p_char);
		if (termBuf->tNotifyFn)
			(*(termBuf->tNotifyFn))(termBuf->tID, TBNOTIFY_NEWCHAR, termBuf->tCurX, termBuf->tCurY,
				termBuf->tCurX, termBuf->tCurY);
	}

	/* note what the following does.. it lets a character hitting the end of
	 * a line to actually "poke" through into an invalid area of the screen.
	 * This is to avoid wrapping until we for sure need to with the reception
	 * of the next character */
	termBuf->tCurX++;

	if (termBuf->tAutoRightMargin) {
		if (termBuf->tCurX > TB_GETXMAX(termBuf)) {
			/* we are off the end of our line. Throw in a wrap-around here first */
			termBuf->tCurX=0;
			/* and do a newline */
			tbLineFeed(termBuf);
		}
	}
	if (termBuf->tNotifyFn)
		(*(termBuf->tNotifyFn))(termBuf->tID, TBNOTIFY_MOVECUR, termBuf->tCurX, termBuf->tCurY, 0L, 0L);
}


/*
 * tbPutChar writes the specified ASCII 8-bit character to the
 * terminal as would be expected, and does all attribute-assignment
 * and scrolling as required.
 */
void tbPutCharReverse(TBTERMBUF *termBuf, unsigned char p_char)
{
	if (termBuf) {
		*TB_GETCHARADDR(termBuf, termBuf->tCurX, termBuf->tCurY) =
			termBuf->tCurrentAttribute|TB_MAKECHAR(p_char);
		if (termBuf->tNotifyFn)
			(*(termBuf->tNotifyFn))(termBuf->tID,TBNOTIFY_NEWCHAR, termBuf->tCurX,termBuf->tCurY,
				termBuf->tCurX, termBuf->tCurY);
	}

	if (termBuf->tCurX>0) {
		/* Cursor moves left one */
		termBuf->tCurX--;
	} else {
		/* line wrap */
		termBuf->tCurX = TB_GETXMAX(termBuf);
		/* and do a newline */
		tbLineFeedUp(termBuf);
	}
	if (termBuf->tNotifyFn)
		(*(termBuf->tNotifyFn))(termBuf->tID, TBNOTIFY_MOVECUR, termBuf->tCurX, termBuf->tCurY, 0L, 0L);
}


void tbSetFColor(TBTERMBUF *termBuf, unsigned long color) {
	if (termBuf) {
		termBuf->tCurrentAttribute = TB_SETATTRFCOLOR(termBuf, color);
	}
}


void tbSetBColor(TBTERMBUF *termBuf, unsigned long color) {
	if (termBuf) {
		termBuf->tCurrentAttribute = TB_SETATTRBCOLOR(termBuf, color);
	}
}


void tbSetEffect(TBTERMBUF *termBuf, unsigned long effect) {
	if (termBuf) {
		termBuf->tCurrentAttribute = TB_SETATTREFFECT(termBuf, effect);
	}
}


void tbSetCursor(TBTERMBUF *termBuf, long pCurX, long pCurY)
{
	if (!termBuf)  return;

	termBuf->tCurX = TB_MIN(TB_MAX(pCurX, 0), TB_GETXMAX(termBuf));
	termBuf->tCurY = TB_MIN(TB_MAX(pCurY, 0), TB_GETYMAX(termBuf));
	if (termBuf->tNotifyFn)
		(*(termBuf->tNotifyFn))(termBuf->tID, TBNOTIFY_MOVECUR, termBuf->tCurX, termBuf->tCurY, 1L, 0L);
}


void tbClearScreen(TBTERMBUF *termBuf)
{
	long  x, y;

	if (!termBuf)  return;

	for (y = 0; y <= TB_GETYMAX(termBuf); y++)
		for (x = 0; x <= TB_GETXMAX(termBuf); x++)
			*TB_GETCHARADDR(termBuf, x, y) = termBuf->tCurrentAttribute|TB_MAKECHAR(0x20);
	if (termBuf->tNotifyFn)
		(*(termBuf->tNotifyFn))(termBuf->tID, TBNOTIFY_NEWCHAR, 0L, 0L,
				TB_GETXMAX(termBuf), TB_GETYMAX(termBuf));
	termBuf->tCurX = termBuf->tCurY = 0;
}


void tbLineFeed(TBTERMBUF *termBuf)
{
	long  i;

	if (!termBuf)  return;

	if (termBuf->tCurY < TB_GETYMAX(termBuf)) {
		/* not at bottom yet.. just increment our cursor Y */
		termBuf->tCurY++;
		i = 1L;
	} else {
		tbScrollUp(termBuf);
		i = 0L;
	}
	if (termBuf->tNotifyFn)
		(*(termBuf->tNotifyFn))(termBuf->tID, TBNOTIFY_MOVECUR,
				termBuf->tCurX, termBuf->tCurY, i, 0L);
}


void tbLineFeedUp(TBTERMBUF *termBuf)
{
	long  i;

	if (!termBuf)  return;

	if (termBuf->tCurY > 0) {
		/* not at top yet.. just increment our cursor Y */
		termBuf->tCurY--;
		i = 1L;
	} else {
		tbScrollDown(termBuf);
		i = 0L;
	}
	if (termBuf->tNotifyFn)
		(*(termBuf->tNotifyFn))(termBuf->tID, TBNOTIFY_MOVECUR,
			  termBuf->tCurX, termBuf->tCurY, i, 0L);
}


void tbScrollDown(TBTERMBUF *termBuf)
{
	long  x, y;

	if (!termBuf)  return;

	/* Copy all data down one line */
	for (y = TB_GETYMAX(termBuf)-1; y >= 0; y--) {
		for (x = 0; x <= TB_GETXMAX(termBuf); x++) {
			*TB_GETCHARADDR(termBuf, x, (y+1)) = *TB_GETCHARADDR(termBuf, x, y);
		}
	}

	/* Clear the top line */
	for (x = 0; x <= TB_GETXMAX(termBuf); x++) {
		*TB_GETCHARADDR(termBuf, x, 0)= termBuf->tCurrentAttribute|TB_MAKECHAR(0x20);
	}
	/* notify */
	if (termBuf->tNotifyFn)
		(*(termBuf->tNotifyFn))(termBuf->tID, TBNOTIFY_SCROLL, -1L, 0L, 0L, 0L);
}


void tbScrollUp(TBTERMBUF *termBuf)
{
	long  i;

	if (!termBuf)  return;

	if (termBuf->tTopY < (termBuf->tBufY-1)) {
		/* not at end of the buffer yet... just move screen down 1 */
		termBuf->tTopY++;
	} else {
		/* at the end of the buffer... wrap screen */
		termBuf->tTopY = 0;
	}
	/* Note, since we've moved our screen, we also have to move our hilight
	 * region (if there is one) */
	termBuf->tHLStartY--;
	termBuf->tHLEndY--;
	if (termBuf->tHLStartY < TB_GETYMIN(termBuf)) {
		termBuf->tHLStartY = TB_GETYMIN(termBuf);
		termBuf->tHLStartX = 0;
	}
	if (termBuf->tHLEndY < TB_GETYMIN(termBuf)) {
		termBuf->tHLEndY = TB_GETYMIN(termBuf);
		termBuf->tHLEndX = 0;
	}

	for (i = TB_GETXMIN(termBuf);i <= TB_GETXMAX(termBuf); i++) {
		*TB_GETCHARADDR(termBuf, i, termBuf->tCurY)= termBuf->tCurrentAttribute|TB_MAKECHAR(0x20);
	}
	if (termBuf->tNotifyFn)
		(*(termBuf->tNotifyFn))(termBuf->tID, TBNOTIFY_SCROLL, 1L, 0L, 0L, 0L);
}


void tbScrollLeft(TBTERMBUF *termBuf)
{
	if (termBuf)
		tbMoveText(termBuf, 1, 0, TB_GETXMAX(termBuf), TB_GETYMAX(termBuf), 0, 0);
}


void tbScrollRight(TBTERMBUF *termBuf) {
	if (termBuf)
		tbMoveText(termBuf, 0, 0, TB_GETXMAX(termBuf), TB_GETYMAX(termBuf), 1, 0);
}


void tbClearToEOL(TBTERMBUF *termBuf, long start)
{
	long  i;

	if (!termBuf)  return;

	if (start < 0)
		start = termBuf->tCurX;
	for (i = start; i <= TB_GETXMAX(termBuf); i++) {
		*TB_GETCHARADDR(termBuf, i, termBuf->tCurY)= termBuf->tCurrentAttribute|TB_MAKECHAR(0x20);
	}
	if (termBuf->tNotifyFn)
		(*(termBuf->tNotifyFn))(termBuf->tID, TBNOTIFY_NEWCHAR,
				start, termBuf->tCurY, TB_GETXMAX(termBuf), termBuf->tCurY);
}


void tbClearLineX(TBTERMBUF *termBuf, long start, long end)
{
	long  i;

	if (!termBuf)  return;
	for (i = start; i <= end; i++) {
		*TB_GETCHARADDR(termBuf, i, termBuf->tCurY)= termBuf->tCurrentAttribute|TB_MAKECHAR(0x20);
	}
	if (termBuf->tNotifyFn)
		(*(termBuf->tNotifyFn))(termBuf->tID, TBNOTIFY_NEWCHAR,
				start, termBuf->tCurY, end, termBuf->tCurY);
}


void tbClearRight(TBTERMBUF *termBuf)
{
	if (!termBuf)  return;
	tbClearLineX(termBuf, termBuf->tCurX, TB_GETXMAX(termBuf));
}


void tbClearLeft(TBTERMBUF *termBuf)
{
	if (!termBuf)  return;
	tbClearLineX(termBuf, 0, termBuf->tCurX);
}


void tbClearLine(TBTERMBUF *termBuf)
{
	if (!termBuf)  return;
	tbClearLineX(termBuf, 0, TB_GETXMAX(termBuf));
}


void tbHTab(TBTERMBUF *termBuf)
{
	int  i;

	if (!termBuf)  return;

	i = termBuf->tCurX;
	do {
		i++;
		if (i > TB_GETXMAX(termBuf))  i = 0;
		if (termBuf->tTab[i])  break;
	} while(i != termBuf->tCurX);
	if ((i <= termBuf->tCurX) && (termBuf->tTab[i])) {
		/* do line down */
		termBuf->tCurX = i;
		tbLineFeed(termBuf);
	} else if (i > termBuf->tCurX) {
		termBuf->tCurX = i;
		(*(termBuf->tNotifyFn))(termBuf->tID, TBNOTIFY_MOVECUR,
					termBuf->tCurX, termBuf->tCurY, 1L, 0L);
	}
}


void tbHTabReverse(TBTERMBUF *termBuf)
{
	int  i;

	if (!termBuf)  return;

	i = termBuf->tCurX;
	do {
		i--;
		if (i < 0)
			i = TB_GETXMAX(termBuf);
		if (termBuf->tTab[i])
			break;
	}	while(i != termBuf->tCurX);
	if ((i >= termBuf->tCurX) && (termBuf->tTab[i])) {
		/* do line up */
		termBuf->tCurX = i;
		tbLineFeedUp(termBuf);
	} else if (i < termBuf->tCurX) {
		termBuf->tCurX = i;
		(*(termBuf->tNotifyFn))(termBuf->tID, TBNOTIFY_MOVECUR,
					termBuf->tCurX, termBuf->tCurY, 1L, 0L);
	}
}


void tbBackSpace(TBTERMBUF *termBuf)
{
	if (!termBuf)  return;

	if (termBuf->tCurX == 0)
		return;
	termBuf->tCurX -= 1;
	if (termBuf->tNotifyFn)
		(*(termBuf->tNotifyFn))(termBuf->tID, TBNOTIFY_MOVECUR,
					termBuf->tCurX, termBuf->tCurY, 1L, 0L);
}


void tbDelete(TBTERMBUF *termBuf)
{
	int i;

	if (!termBuf)  return;

	for (i = termBuf->tCurX+1; i <= TB_GETXMAX(termBuf); i++) {
		*TB_GETCHARADDR(termBuf, i-1, termBuf->tCurY) =
		*TB_GETCHARADDR(termBuf, i, termBuf->tCurY);
	}
	*TB_GETCHARADDR(termBuf, TB_GETXMAX(termBuf), termBuf->tCurY) = termBuf->tCurrentAttribute|TB_MAKECHAR(0x20);
	if (termBuf->tNotifyFn) {
		(*(termBuf->tNotifyFn))(termBuf->tID,TBNOTIFY_NEWCHAR,
					termBuf->tCurX, termBuf->tCurY, TB_GETXMAX(termBuf), termBuf->tCurY);
	}
}


void tbCurUp(TBTERMBUF *termBuf)
{
	if (!termBuf)  return;

	if (termBuf->tCurY > 0) {
		termBuf->tCurY--;
		(*(termBuf->tNotifyFn))(termBuf->tID, TBNOTIFY_MOVECUR,
					termBuf->tCurX, termBuf->tCurY, 1L, 0L);
	}
}


void tbCurDown(TBTERMBUF *termBuf)
{
	if (!termBuf)  return;

	if (termBuf->tCurY < TB_GETYMAX(termBuf)) {
		termBuf->tCurY++;
		(*(termBuf->tNotifyFn))(termBuf->tID, TBNOTIFY_MOVECUR,
					termBuf->tCurX, termBuf->tCurY, 1L, 0L);
	}
}


void tbCurRight(TBTERMBUF *termBuf)
{
	if (!termBuf)  return;

	if (termBuf->tCurX < TB_GETXMAX(termBuf)) {
		termBuf->tCurX++;
		(*(termBuf->tNotifyFn))(termBuf->tID, TBNOTIFY_MOVECUR,
					termBuf->tCurX, termBuf->tCurY, 1L, 0L);
	}
}


void tbCurLeft(TBTERMBUF *termBuf)
{
	if (!termBuf)  return;

	if (termBuf->tCurX > 0) {
		termBuf->tCurX--;
		(*(termBuf->tNotifyFn))(termBuf->tID, TBNOTIFY_MOVECUR,
					termBuf->tCurX, termBuf->tCurY, 1L, 0L);
	}
}


void tbInsertChar(TBTERMBUF *termBuf, int spaces)
{
	long  i;

	if (!termBuf)  return;

	if (spaces < 0)  spaces = 1;
	if (spaces > TB_GETXMAX(termBuf)-termBuf->tCurX) {
		tbClearRight(termBuf);
	} else {
		for (i = TB_GETXMAX(termBuf) - spaces; i >= termBuf->tCurX; i--) {
			*TB_GETCHARADDR(termBuf, i + spaces, termBuf->tCurY) =
			*TB_GETCHARADDR(termBuf, i, termBuf->tCurY);
		}
		for (i = termBuf->tCurX; i < termBuf->tCurX + spaces; i++)
			*TB_GETCHARADDR(termBuf, i, termBuf->tCurY)= termBuf->tCurrentAttribute|TB_MAKECHAR(0x20);

		if (termBuf->tNotifyFn)
			(*(termBuf->tNotifyFn))(termBuf->tID, TBNOTIFY_NEWCHAR,
						termBuf->tCurX, termBuf->tCurY, TB_GETXMAX(termBuf), termBuf->tCurY);
	}
}


void tbClearBelow(TBTERMBUF *termBuf)
{
	long  x, y;

	if (!termBuf)  return;

	for (y = termBuf->tCurY; y <= TB_GETYMAX(termBuf); y++) {
		if (y == termBuf->tCurY)
			x = termBuf->tCurX;
		else
			x = 0;

		for (; x <= TB_GETXMAX(termBuf); x++)
			*TB_GETCHARADDR(termBuf, x, y) = termBuf->tCurrentAttribute|TB_MAKECHAR(0x20);
	}
	if (termBuf->tNotifyFn)
		(*(termBuf->tNotifyFn))(termBuf->tID, TBNOTIFY_NEWCHAR,
					0L, termBuf->tCurY, TB_GETXMAX(termBuf), TB_GETYMAX(termBuf));
}


void tbClearAbove(TBTERMBUF *termBuf)
{
	long  x, y;

	if (!termBuf)  return;

	for (y = 0; y <= termBuf->tCurY; y++) {
		if (y == termBuf->tCurY)
			x = termBuf->tCurX;
		else
			x = TB_GETXMAX(termBuf);
		for (; x >= 0; x--)
			*TB_GETCHARADDR(termBuf, x, y) = termBuf->tCurrentAttribute|TB_MAKECHAR(0x20);
	}
	if (termBuf->tNotifyFn)
		(*(termBuf->tNotifyFn))(termBuf->tID, TBNOTIFY_NEWCHAR,
					0L, 0L, TB_GETXMAX(termBuf), termBuf->tCurY);
}


void tbInsertLine(TBTERMBUF *termBuf, long numLines)
{
	long  i;

	if (!termBuf) return;

	for (i = 0; i < numLines; i++) {
		tbMoveText(termBuf, 0, termBuf->tCurY, TB_GETXMAX(termBuf), TB_GETYMAX(termBuf), 0, termBuf->tCurY+1);
	}
}


void tbDeleteLine(TBTERMBUF *termBuf, long numLines)
{
	long  i;

	if (!termBuf) return;

	for (i = 0; i < numLines; i++) {
		tbMoveText(termBuf,	0, termBuf->tCurY+1, TB_GETXMAX(termBuf), TB_GETYMAX(termBuf), 0, termBuf->tCurY);
	}
}


/*
 * Moves text from one place to another on the screen. It blanks the
 * source area with 0x20 (spaces) in the current background color
 * note this proc DOES NOT pay ANY attention to the current scrolling region
 * Furthermore, it can handle copying from and to regions outside the current screen
 */
void tbMoveText(TBTERMBUF *termBuf,long p_startX,long p_startY,
	long p_endX,long p_endY,long p_destX,long p_destY)
{
	long l_x,l_y,l_dstx,l_dsty,l_OffsetX,l_OffsetY;
	unsigned long l_Char,*l_pChar;
	long l_startX,l_startY,l_endX,l_endY,l_stepX,l_stepY;

	if (!termBuf)  return;

	/* let's make sure we go from top-left to bottom right */
	if (p_endX < p_startX) {
		l_x = p_startX;
		p_startX = p_endX;
		p_endX = l_x;
	}
	if (p_endY < p_startY) {
		l_y = p_startY;
		p_startY = p_endY;
		p_endX = l_y;
	}

	/* First, let's figure out the order in which we will be copying */
	if (p_startX < p_destX) {
		l_startX = p_endX;
		l_endX = p_startX;
		l_stepX = -1;
	} else {
		l_startX = p_startX;
		l_endX = p_endX;
		l_stepX = 1;
	}
	l_OffsetX = p_destX - p_startX;
	if (p_startY < p_destY) {
		l_startY = p_endY;
		l_endY = p_startY;
		l_stepY = -1;
	} else {
		l_startY = p_startY;
		l_endY = p_endY;
		l_stepY = 1;
	}
	l_OffsetY = p_destY - p_startY;

	for (l_y = l_startY; l_y != (l_endY+l_stepY); l_y += l_stepY) {
		for (l_x = l_startX; l_x != (l_endX+l_stepX); l_x += l_stepX) {
			/* first, let's get & clear our source */
			if ((l_y >= 0) && (l_y <= TB_GETYMAX(termBuf)) &&
						(l_x >= 0) && (l_x <= TB_GETXMAX(termBuf))) {
				l_pChar = TB_GETCHARADDR(termBuf, l_x, l_y);
				l_Char = *l_pChar;
				*l_pChar = termBuf->tCurrentAttribute|TB_MAKECHAR(0x20);
			} else {
				l_Char = termBuf->tCurrentAttribute|TB_MAKECHAR(0x20);
			}
			/* next, let's save this char in it's destination spot */
			l_dstx = l_x+l_OffsetX;
			l_dsty = l_y+l_OffsetY;
			if ((l_dsty >= 0) && (l_dsty <= TB_GETYMAX(termBuf))&&
						(l_dstx >= 0) && (l_dstx <= TB_GETXMAX(termBuf))) {
				*TB_GETCHARADDR(termBuf,l_dstx,l_dsty) = l_Char;
			}
		}
	}
	if (termBuf->tNotifyFn)
		(*(termBuf->tNotifyFn))(termBuf->tID, TBNOTIFY_NEWCHAR,
				(p_startX < p_destX) ? p_startX : p_destX,
				(p_startY < p_destY) ? p_startY : p_destY,
				(p_startX < p_destX) ? p_endX+l_OffsetX : p_endX,
				(p_startY < p_destY) ? p_endY+l_OffsetY : p_endY);
}


void tbCreateUnderscorePen(TBTERMBUF *termBuf, int fontHeight)
{
	int  i, width;

	if (!termBuf)  return;

	width = (fontHeight+5)/10;
	if (!width)
		width = 1;
	for (i = 0; i < (1<<TB_COLORDEPTH); i++) {
		if ((termBuf->tColorPen)[i]) {
			DeleteObject((termBuf->tColorPen)[i]);
		}
		(termBuf->tColorPen)[i] = CreatePen(PS_SOLID, width, termBuf->tColorBank[i]);
	}
}


void tbCursorSave(TBTERMBUF *termBuf)
{
	if (!termBuf)  return;

	termBuf->tSaveCurrentAttribute = termBuf->tCurrentAttribute;
	termBuf->tSaveCurX = termBuf->tCurX;
	termBuf->tSaveCurY = termBuf->tCurY;
}


void tbCursorRestore(TBTERMBUF *termBuf)
{
	if (!termBuf)  return;

	termBuf->tCurrentAttribute = termBuf->tSaveCurrentAttribute;
	termBuf->tCurX = termBuf->tSaveCurX;
	termBuf->tCurY = termBuf->tSaveCurY;
}


void tbTabSet(TBTERMBUF *termBuf, int type, int x)
{
	if (!termBuf)  return;

	if ((x >= 0) && (x <= (TB_GETXMAX(termBuf)))) {
		termBuf->tTab[x] = type;
	} else {
		termBuf->tTab[termBuf->tCurX] = type;
	}
}


void tbTabClearAll(TBTERMBUF *termBuf)
{
	int  i;

	if (!termBuf)  return;

	for (i = 0; i <= TB_GETXMAX(termBuf); i++) {
		termBuf->tTab[i] = TBTAB_NONE;
	}
}


void tbStartHilight(TBTERMBUF *termBuf, int x, int y)
{
	if (!termBuf)  return;

	if (y > TB_GETYMAX(termBuf)) {
		y = TB_GETYMAX(termBuf);
		x = TB_GETXMAX(termBuf);
	} else if (y < TB_GETYMIN(termBuf)) {
		y = TB_GETYMIN(termBuf);
		x = TB_GETXMIN(termBuf);
	} else if (x > TB_GETXMAX(termBuf)) {
		x = TB_GETXMAX(termBuf);
	} else if (x < TB_GETXMIN(termBuf)) {
		x = TB_GETXMIN(termBuf);
	}
	termBuf->tHLStartX = termBuf->tHLEndX = x;
	termBuf->tHLStartY = termBuf->tHLEndY = y;
}


void tbUpdateHilight(TBTERMBUF *termBuf, int x, int y)
{
	if (!termBuf)  return;

	if (y > TB_GETYMAX(termBuf)) {
		y = TB_GETYMAX(termBuf);
		x = TB_GETXMAX(termBuf)+1;
	} else if (y <TB_GETYMIN(termBuf)) {
		y = TB_GETYMIN(termBuf);
		x = TB_GETXMIN(termBuf);
	} else if (x >TB_GETXMAX(termBuf)+1) {
		x = TB_GETXMAX(termBuf)+1;
	} else if (x <TB_GETXMIN(termBuf)) {
		x = TB_GETXMIN(termBuf);
	}
	termBuf->tHLEndX= x;
	termBuf->tHLEndY= y;
}


void tbClearHilight(TBTERMBUF *termBuf)
{
	if (!termBuf)  return;

	termBuf->tHLStartX = termBuf->tHLStartY = 0L;
	termBuf->tHLEndX   = termBuf->tHLEndY   = 0L;
}


BOOL tbIsHilight(TBTERMBUF *termBuf, int x, int y)
{
	if (!termBuf)  return FALSE;

	if (termBuf->tHLStartY < termBuf->tHLEndY) {
		/* we go from start, FORWARD to our end */
		if ((y > termBuf->tHLStartY) && (y<termBuf->tHLEndY))
			return TRUE;
		if (y == termBuf->tHLStartY) {
			/* on starting line */
			if (x >= termBuf->tHLStartX)
				return TRUE;
		} else if (y == termBuf->tHLEndY) {
			/* on ending line */
			if (x < termBuf->tHLEndX)
				return TRUE;
		}
		return FALSE;
	} else if (termBuf->tHLStartY > termBuf->tHLEndY) {
		/* we go from start, BACKWARD to our end */
		if ((y > termBuf->tHLEndY) && (y < termBuf->tHLStartY))
			return TRUE;
		if (y == termBuf->tHLEndY) {
			/* on starting line */
			if (x >= termBuf->tHLEndX)
				return TRUE;
		} else if (y == termBuf->tHLStartY) {
			/* on ending line */
			if (x < termBuf->tHLStartX)
				return TRUE;
		}
		return FALSE;
	} else if (termBuf->tHLStartX < termBuf->tHLEndX) {
		/* we go from start, FORWARD to our end */
		/* single line! */
		if ((y == termBuf->tHLStartY) &&
				(x >= termBuf->tHLStartX) &&
				(x < termBuf->tHLEndX) )
			return TRUE;
		return FALSE;
	} else if (termBuf->tHLStartX > termBuf->tHLEndX) {
		/* we go from start, BACKWARD to our end */
		/* single line! */
		if ((y == termBuf->tHLStartY) &&
				(x >= termBuf->tHLEndX) &&
				(x < termBuf->tHLStartX) )
			return TRUE;
		return FALSE;
	}
	/* no hilight */
	return FALSE;
}


/*
 * Returns the size, in bytes, the hilighted region would be expected to consume
 */
int tbHilightSize(TBTERMBUF *termBuf)
{
	int  total;

	if (!termBuf)  return 0;

	if (termBuf->tHLStartY < termBuf->tHLEndY) {
		/* we go from start, FORWARD to our end */
		/* first, add in our full, margin-to-margin lines */
		total = (termBuf->tHLEndY-termBuf->tHLStartY-1) * (TB_GETXMAX(termBuf) + 1 + 2);
		/* next, our start line chunk */
		total += (TB_GETXMAX(termBuf) - termBuf->tHLStartX+1);
		/* finally, our end line chunk */
		total += termBuf->tHLEndX;
		/* and allow for a few extra CRLF's */
		total += 6;
	} else if (termBuf->tHLStartY > termBuf->tHLEndY) {
		/* we go from start, BACKWARD to our end */
		/* first, add in our full, margin-to-margin lines */
		total = (termBuf->tHLStartY - termBuf->tHLEndY-1) * (TB_GETXMAX(termBuf) + 1 + 2);
		/* next, our start line chunk */
		total += (TB_GETXMAX(termBuf)-termBuf->tHLEndX+1);
		/* finally, our end line chunk */
		total += termBuf->tHLStartX;
		/* and allow for a few extra CRLF's */
		total += 6;
	} else if (termBuf->tHLStartX < termBuf->tHLEndX) {
		/* we go from start, FORWARD to our end */
		/* single line! */
		total = termBuf->tHLEndX-termBuf->tHLStartX;
		/* and allow for a few extra CRLF's */
		total += 6;
	} else if (termBuf->tHLStartX > termBuf->tHLEndX) {
		/* we go from start, BACKWARD to our end */
		/* single line! */
		total = termBuf->tHLStartX-termBuf->tHLEndX;
		/* and allow for a few extra CRLF's */
		total += 6;
	} else {
		/* no hilight */
		total = 0; /* a single 0x00 */
	}

	return total;
}


/*
 * Copies hilighted region into buffer, inserting CRLF at the ends of lines
 */
void tbCopyHilight(TBTERMBUF *termBuf, unsigned char *pBuf)
{
	int  sx, sy, ex, ey, x, y;
	int  c, countEndX;
	unsigned char *p;

	if ((termBuf) && (pBuf)) {
		if (termBuf->tHLStartY < termBuf->tHLEndY) {
			/* we go from start, FORWARD to our end */
			sx = termBuf->tHLStartX;
			sy = termBuf->tHLStartY;
			ex = termBuf->tHLEndX;
			ey = termBuf->tHLEndY;
		} else if (termBuf->tHLStartY > termBuf->tHLEndY) {
			/* we go from start, BACKWARD to our end */
			sx = termBuf->tHLEndX;
			sy = termBuf->tHLEndY;
			ex = termBuf->tHLStartX;
			ey = termBuf->tHLStartY;
		} else if (termBuf->tHLStartX < termBuf->tHLEndX) {
			/* we go from start, FORWARD to our end */
			sx = termBuf->tHLStartX;
			sy = termBuf->tHLStartY;
			ex = termBuf->tHLEndX;
			ey = termBuf->tHLEndY;
		} else if (termBuf->tHLStartX > termBuf->tHLEndX) {
			/* we go from start, BACKWARD to our end */
			sx = termBuf->tHLEndX;
			sy = termBuf->tHLEndY;
			ex = termBuf->tHLStartX;
			ey = termBuf->tHLStartY;
		} else {
			/* no hilight */
			pBuf[0]=0;
			return;
		}
		p = pBuf;
		for (y = sy; y <= ey; y++) {
			if (y == sy)
				x = sx;
			else
				x = 0;

			if (y == ey)
				countEndX = ex;
			else
				countEndX = TB_GETXMAX(termBuf)+1;

			for (; x < countEndX; x++) {
				/* copy this character */
				c = TB_GETCHAR(termBuf, x, y);
				*p = (unsigned char)TB_EXTCHAR(c);
				p++;
			}
			/* As a courtesy, let's remove any trailing spaces */
			p--;
			while((p >= pBuf) && (*p == ' ')) {
				p--;
			}
			p++;
			/* end of line -- add CRLF */
			if (x > TB_GETXMAX(termBuf)) {
				*p++ = 0x0D;
				*p++ = 0x0A;
			}
		}
		*p = '\0';
	}

	return;
}

